# environment.py

import ast
import json
import re
import logging
from django.template import engines
from integrations.jinja import getSecretByKey

logger = logging.getLogger('connectors')


def expand(env, obj, workflow=None):
    """
    Expands an object using dynamic variables specified in the given environment

   :param dict env: The environment containing available dynamic variables
   :param Any obj: The object to expand

   :return: The expanded object
   :rtype: Any
    """
    if not env:
        return obj
    return _expand(env, obj, workflow)


def _expand(env, obj, workflow=None):
    """
    Expands an object using dynamic variables specified in the given environment

   :param dict env: The environment containing avaiable dynamic variables
   :param Any obj: The object to expand

   :return: The expanded object
   :rtype: Any
    """
    obj_type = type(obj).__name__
    if obj_type == 'dict':
        ret = {}
        for key, value in obj.items():
            ret[key] = _expand(env, value)
        return ret
    elif obj_type == 'list':
        return [_expand(env, list_item) for list_item in obj]
    elif obj_type == 'str':
        return _expand_string(env, obj, workflow)

    return obj


jinja2 = engines['jinja2']


def _expand_string(env, string, workflow=None):
    """
    Expands the string using jinja, with env as the context object.

   :param dict env: The environment containing available dynamic variables
   :param str string: The string to expand via jinja

   :return: The templated string
   :rtype: Any
    """
    env['request'] = env.get('request', {})
    context = {
        'vars': env,
        'request': env.get('request'),
        'files': env.get('files', {}),
    }

    if '{{secret_store.' in string.replace(' ', ''):
        pattern = r"{{secret_store.+?}}"
        secret_store = {}
        for each_string in re.findall(pattern, string):
            arg1, arg2 = _eval_args(each_string)
            if arg1 and arg2:
                lookup_key = '{0}.{1}'.format(arg1, arg2)
                if arg1 in secret_store:
                    secret_store[arg1].update({arg2: getSecretByKey(lookup_key)})
                else:
                    secret_store.update({arg1: {arg2: getSecretByKey(lookup_key)}})
            elif arg1 and not arg2:
                lookup_key = arg1
                secret_store.update({'{0}'.format(lookup_key): getSecretByKey(lookup_key)})

        context.update({'secret_store': secret_store})


    template = jinja2.from_string(string)
    ret = template.render(context=context)
    try:
        ret = json.loads(ret, strict=False)
        return ret
    except Exception:
        pass
    try:
        ret = ast.literal_eval(ret)
        return ret
    except (SyntaxError, ValueError):
        pass

    return ret

def _eval_args(string):
    tmp = (string.replace(' ', '').replace('{{', '').replace('}}', '').split('|')[0]).\
        replace('[', '.').replace(']', '.').replace('..', '.').split('.')
    arg1 = tmp[1]
    arg2 = None
    if len(tmp) > 2:
        arg2 = tmp[2]
    try:
        arg1 = json.loads(arg1)
    except:
        pass
    try:
        arg2 = json.loads(arg2)
    except:
        pass
    return arg1, arg2
