import copy

from django.core.management import BaseCommand
from connectors.serializers import ConnectorDetailSerializer
from connectors.models import Connector
from connectors.views import update_connector, install_connector_packages


class Command(BaseCommand):
    help = (
        "Helps to update connector metadata in the database from the its latest info.json."
    )
    missing_args_message = "You must specify either the connector id/name-version or --migrate " \
                           "for reimporting all connectors."

    def add_arguments(self, parser):
        parser.add_argument('-c', '--conn_id', type=str, help='connector id')
        parser.add_argument('-n', '--name', type=str, help='connector name')
        parser.add_argument('-cv', '--conn_version', type=str, help='connector version')
        parser.add_argument(
            '-m'
            '--migrate',
            action='store_true',
            dest='migrate',
            help='Re-import connector info',
        )

    def handle(self, **options):

        try:
            conn_id = options.get('conn_id', None)
            name = options.get('name', None)
            version = options.get('conn_version', None)
            migrate = options.get('migrate', True)

            all_connnectors = []
            if conn_id:
                all_connnectors = Connector.objects.filter(id=conn_id)
            elif name and version:
                all_connnectors = Connector.objects.filter(name=name, version=version)
            elif migrate:
                all_connnectors = Connector.objects.all()
            else:
                self.stdout.write(self.style.ERROR(
                    'When not migrating connector id or name and version are required'))
            if all_connnectors:
                for connector in all_connnectors:
                    self.stdout.write('Re-Importing connector {0} v{1}'.format(connector.name, connector.version))
                    config = {}
                    if migrate:
                        config = connector.configuration_old
                    if (migrate and not connector.migrate) or not migrate:
                        serializer = ConnectorDetailSerializer(connector)
                        connector_object = serializer.data
                        connector_object_copy = copy.deepcopy(connector_object)
                        connector.delete()
                        try:
                            update_connector(connector=connector_object, configurations = config)
                            self.stdout.write(self.style.SUCCESS(
                                'Successfully re-imported connector {0} v{1}'.format(connector.name, connector.version)))

                            self.stdout.write('Installing dependant packages for connector {0} v{1}'.format(
                                connector.name, connector.version))
                            install_connector_packages(connector.name,connector.version)
                            self.stdout.write(self.style.SUCCESS(
                                'Successfully installed all the packages for connector {0} v{1}'.format(connector.name, connector.version)))

                        except Exception as e:
                            self.stdout.write(self.style.ERROR(
                                'Error occurred while updating the connector data ERROR :: {0}'.format(str(e))))
                            serializer = ConnectorDetailSerializer(data=connector_object_copy)
                            serializer.is_valid(raise_exception=True)
                            serializer.save()
                    elif migrate:
                        self.stdout.write(self.style.SUCCESS('Connector {0} v{1} is already migrated'.format(
                            connector.name, connector.version)))
        except Exception as e:
            self.stdout.write(self.style.ERROR('Error occurred while migrating the connector data ERROR :: {0}'.format(str(e))))
