import os
import inspect
import importlib
import logging
import subprocess
import base64
import shutil, psutil
import uuid
import json
from django.conf import settings
from django.db.models import Q
from connectors.core.base_connector import Connector, ConnectorError, STATE_AVAILABLE
from connectors.models import Connector as ConnectorModel
from connectors.models import Configuration as ConfigurationModel
from connectors.models import Operation as OperationModel
from connectors.serializers import ConnectorDetailSerializer, ConnectorConfigurationSerializer, \
    ConnectorOperationSerializer
from connectors.environment import _expand
from integrations.crudhub import make_request, make_file_upload_request

logger = logging.getLogger('connectors')


def get_connector(item, version, info_json=None):
    try:
        version = '%s' % (version.replace('.', '_'))
        connector_name = '%s_%s' % (item, version)
        connector_module = 'connectors.%s' % connector_name
        connector_module = importlib.import_module('%s.connector' % connector_module)
        importlib.reload(connector_module)
        for name, obj in inspect.getmembers(connector_module):
            if inspect.isclass(obj) and issubclass(obj, Connector) and not inspect.isabstract(obj):
                try:
                    connector_instance = obj(info_json=info_json)
                except:
                    connector_instance = obj()
                return {'instance': connector_instance}
    except Exception as e:
        return {'instance': None, 'message': str(e)}


def get_connector_instance(conn_name, conn_version, conn_obj=None):
    info_json = {}
    if conn_obj:
        connector_serializer = ConnectorDetailSerializer(conn_obj)
        info_json = connector_serializer.data
    connector = get_connector(conn_name, conn_version, info_json)
    if connector.get('instance'):
        return connector.get('instance')
    else:
        logger.error(
            'Exception occurred while getting the connector instance ERROR :: {0}'.format(connector.get('message')))
        raise ConnectorError(connector.get('message'))


def invalidate_cache():
    if importlib is not None:
        if hasattr(importlib, "invalidate_caches"):
            importlib.invalidate_caches()


def get_connector_version_or_latest(name, version, status='Completed'):
    if status:
        connector_instance = ConnectorModel.objects.filter(name=name, version=version, status='Completed')
    else:
        connector_instance = ConnectorModel.objects.filter(name=name, version=version)
    if not connector_instance.exists():
        version_latest = None
        if status:
            connectors = ConnectorModel.objects.filter(name=name, status='Completed')
        else:
            connectors = ConnectorModel.objects.filter(name=name)
        for each_connector in connectors:
            version_current = each_connector.version
            if not version_latest:
                version_latest = version_current
            else:
                if tuple(map(int, (version_current.split(".")))) > tuple(map(int, (version_latest.split(".")))):
                    version_latest = version_current
        if version_latest:
            version = version_latest
    return version


def get_connector_path(name, version):
    version = '%s' % (version.replace('.', '_'))
    return os.path.join(settings.CONNECTORS_DIR, ('%s_%s' % (name, version)))


def call_connectors_on_app_start_func():
    try:
        if is_active_primary()[0]:
            from connectors.models import Connector
            connectors = Connector.objects.filter(status='Completed')
            for connector in connectors:
                try:
                    logger.info('Calling on_app_start() for connector %s', connector.name)
                    conn_instance = get_connector_instance(connector.name, connector.version)
                    config = get_configuration(conn_id=connector.id, decrypt=True,
                                               config_schema=connector.config_schema, convert_to_dict=True)
                    conn_instance.on_app_start(config, connector.active)
                except Exception as e:
                    logger.exception('Error while calling connector %s on_app_start function: %s', connector.name,
                                     str(e))
    except:
        pass


def manage_password(data, action):
    if not hasattr(settings, 'ENCRYPTION') or settings.ENCRYPTION:
        if action == 'decrypt':
            if data.endswith('I3dmcn23@KlS2#!ck'):
                output = subprocess.check_output(['/opt/cyops-auth/.env/bin/python',
                                                  '/opt/cyops/configs/scripts/manage_passwords.py',
                                                  '--decrypt',
                                                  data,
                                                  'jQp3(7@jod#j38d1']
                                                 ).strip().decode()
                if not output.startswith('Password:'):
                    raise Exception('DB Password decrypt failed with Error: %s for Password : %s' % output)
                else:
                    return output[len('Password:'):]

            else:
                return data

        elif action == 'encrypt':
            if data and not data.endswith('I3dmcn23@KlS2#!ck'):
                output = subprocess.check_output(['/opt/cyops-auth/.env/bin/python',
                                                  '/opt/cyops/configs/scripts/manage_passwords.py',
                                                  '--encrypt',
                                                  data]
                                                 ).strip().decode()
                return output
            else:
                return data
    return data


def encrypt_password(config_schema, config):
    fields_with_type_pwd = [each_field.get('name') for each_field in config_schema.get('fields', []) if
                            each_field.get('type') == 'password']
    for field in fields_with_type_pwd:
        if config.get(field):
            config[field] = manage_password(data=config[field], action='encrypt')


def decrypt_password(config, config_schema):
    fields_with_type_pwd = [each_field.get('name') for each_field in config_schema.get('fields', []) if
                            each_field.get('type') == 'password']
    for field in fields_with_type_pwd:
        if config.get(field):
            config[field] = manage_password(data=config[field], action='decrypt')


def omit_password(result):
    configurations = result.get('configuration', [])
    config_schema = result.get('config_schema', {})
    fields_with_type_pwd = [each_field.get('name') for each_field in config_schema.get('fields', []) if
                            each_field.get('type') == 'password']
    if isinstance(configurations, list):
        for v in configurations:
            config = v.get('config', {})
            for field in fields_with_type_pwd:
                if field in config:
                    config[field] = 'NULL'
    else:
        config = configurations.get('config', {})
        for field in fields_with_type_pwd:
            if field in config:
                config[field] = 'NULL'


def update_pwd_field(new_config, old_config, config_schema):
    fields_with_type_pwd = [each_field.get('name') for each_field in config_schema.get('fields', []) if
                            each_field.get('type') == 'password']

    for field in fields_with_type_pwd:
        if new_config.get(field) and new_config.get(field) == 'NULL':
            new_config[field] = manage_password(data=old_config[field], action='decrypt')


def on_connector_delete(conn_name, conn_version, config):
    if is_active_primary()[0]:
        get_connector_instance(conn_name, conn_version).teardown(config)


def on_active_status_update(conn_name, conn_version, old_status, new_status, config):
    if old_status != new_status and is_active_primary()[0]:
        conn_instance = get_connector_instance(conn_name, conn_version)
        if new_status:
            conn_instance.on_activate(config)
        else:
            conn_instance.on_deactivate(config)


def process_images(data):
    icon_small, icon_large = '', ''
    try:
        with open(os.path.join(
                get_connector_path(data.get('name'), data.get('version')),
                'images', data.get('icon_small_name')), 'rb') as image_file:
            icon_small = 'data:image/jpeg;base64,' + (
                base64.b64encode(image_file.read())).decode()
    except Exception as e:
        logger.warn('Error while saving icon_small jpeg image %s', str(e))
    try:
        with open(os.path.join(
                get_connector_path(data.get('name'), data.get('version')),
                'images', data.get('icon_large_name')), 'rb') as image_file:
            icon_large = 'data:image/jpeg;base64,' + (
                base64.b64encode(image_file.read())).decode()
    except Exception as e:
        logger.warn('Error while saving icon_small jpeg image %s', str(e))

    return icon_small, icon_large


def get_parsed_config(config):
    '''
    This function will parse if config is in the form of jinja template.
    :param config:
    :return:
    '''
    try:
        return {key: _expand({}, value) for key, value in config.items()}
    except Exception as e:
        logger.error('Error while evaluating jinja template for config %s', str(e))
        raise e


def get_previous_version_config(info, replace=True):
    # this will be used to import the previous version config in new version if config schema is same
    try:
        latest_version = get_connector_version_or_latest(info.get('name'),
                                                         info.get('version'))
        if latest_version and ConnectorModel.objects.filter(name=info.get('name'), version=latest_version,
                                                            status='Completed').exists():
            connector_detail = \
                ConnectorModel.objects.filter(name=info.get('name'), version=latest_version).first()
            connector_serializer = ConnectorDetailSerializer(connector_detail)
            connector_detail = connector_serializer.data
            if bool(info.get('configuration')):
                configuration = connector_detail.get('configuration')
                return import_prev_config(info.get('configuration'), connector_detail.get('config_schema'),
                                          configuration, connector_detail.get('id'), replace)
    except Exception as e:
        logger.error("Error occurred while importing the previous configuration ERROR :: {0}".format(str(e)))
    return {}


def symlink_system_connector(connector_name, connector_version):
    create_symlink = True
    connectors = ConnectorModel.objects.filter(name=connector_name)
    if len(connectors) > 0:
        for each_connector in connectors:
            version_current = each_connector.version
            if tuple(map(int, (version_current.split(".")))) > tuple(map(int, (connector_version.split(".")))):
                create_symlink = False
                break

    if create_symlink:
        symlink_dir = os.path.join(settings.CONNECTORS_DIR, connector_name)
        # remove previous symlink if any
        try:
            os.unlink(symlink_dir)
        except:
            logger.info('No previous symlink for the connector: %s' % symlink_dir)
        # for upgraded system from cyops < 4.10.3, this could be actual dir
        try:
            shutil.rmtree(symlink_dir)
        except:
            logger.info('No previous directory for connector: %s' % symlink_dir)
        # create the new symlink
        os.symlink(get_connector_path(connector_name, connector_version), symlink_dir)


def get_configuration(conn_id=None, config_id=None, decrypt=False, config_schema={}, convert_to_dict=False):
    try:
        config_object = []
        many = False

        if not bool(config_schema and config_schema.get('fields', [])) and config_id:
            return {}
        elif not config_schema:
            return []

        if config_id:
            config_instance = ConfigurationModel.objects.filter(
                Q(connector=conn_id, config_id=config_id) | Q(connector=conn_id, name=config_id)).first()
            if not config_instance:
                config_instance = ConfigurationModel.objects.filter(connector=conn_id, default=True).first()
            if not config_instance:
                message = 'Could not find a configuration matching the id {0} or the default configuration for the connector'
                logger.error(message.format(config_id))
                raise ConnectorError(message.format(config_id))
            if config_instance.status == 0:
                message = "The configuration {0} is only partially setup. Cannot be used till the configuration is complete"
                logger.error(message.format(config_instance.name))
                raise ConnectorError(message.format(config_instance.name))
        else:
            many = True
            config_instance = ConfigurationModel.objects.filter(connector=conn_id)

        if config_instance:
            serializer = ConnectorConfigurationSerializer(config_instance, many=many)
            config_object = serializer.data

        if isinstance(config_object, list):
            dict_config = {}
            for v in config_object:
                config = v.get('config', {})
                if decrypt:
                    decrypt_password(config, config_schema)
                config = get_parsed_config(config)
                config['config_id'] = v.get('config_id')
                config['name'] = v.get('name')
                config['status'] = v.get('status', 0)
                if convert_to_dict:
                    dict_config[config['config_id']] = config
            if convert_to_dict:
                config_object = dict_config

        else:
            config = config_object.get('config', {})
            if decrypt:
                decrypt_password(config, config_schema)
            config = get_parsed_config(config)
            config['config_id'] = config_object.get('config_id')
            config['name'] = config_object.get('name')
            config['status'] = config_object.get('status', 0)
            config_object = config

        return config_object

    except ConnectorError as e:
        raise ConnectorError(str(e))
    except Exception as e:
        logger.exception(
            "Error occurred while retrieving configuration with config id : {0} ERROR :: {1} ".format(config_id,
                                                                                                      str(e)))
        raise ConnectorError("Error occurred while retrieving configuration with config id : {0} ".format(config_id))


def get_operation(name=None, conn_id=None):
    try:
        operation_instance = OperationModel.objects.get(connector=conn_id, operation=name, enabled=True)
        serializer = ConnectorOperationSerializer(operation_instance)
        return serializer.data
    except Exception as e:
        raise ConnectorError(
            "Error occurred while retrieving the connector operation {0} ERROR :: {1}".format(name, str(e)))


def config_type_check(name, type, schema):
    for param in schema:
        if name == param.get('name', ''):
            if not type == param.get('type', ''):
                return False
            break
    return True


def get_nested_config_schema(new_onchange_value, param_value, param_name, old_schema):
    new_nested_config = new_onchange_value.get(str(new_onchange_value),
                                               new_onchange_value.get(str(param_value).lower(), []))
    new_nested_schema = {"fields": new_nested_config}
    old_nested_config = []
    for old_schema_value in old_schema:
        if old_schema_value.get('name', None) == param_name:
            old_onchange_value = old_schema_value.get('onchange', None)
            if old_onchange_value:
                old_nested_config = old_onchange_value.get(str(param_value),
                                                           old_onchange_value.get(str(param_value).lower(), []))
            break
    old_nested_schema = {"fields": old_nested_config}
    return new_nested_schema, old_nested_schema


def import_prev_config(new_schema, old_schema, configuration, conn_id, replace=True):
    new_schema = new_schema.get('fields', {})
    old_schema = old_schema.get('fields', {})
    for config_object in configuration:
        valid_config = {}
        config = config_object.get('config', {})
        config_status = 1
        for param in new_schema:
            param_name = param.get('name', '')
            param_value = config.get(param_name, None)
            same_type = config_type_check(param_name, param.get('type'), old_schema)
            new_onchange_value = param.get('onchange', None)
            if new_onchange_value:
                new_nested_schema, old_nested_schema = get_nested_config_schema(new_onchange_value, param_value,
                                                                                param_name, old_schema)
                valid_nested_config = import_prev_config(new_nested_schema, old_nested_schema, [config_object], conn_id)
                config_status = valid_nested_config[0].get('status')
                valid_config.update(valid_nested_config[0].get('config'))
            if param_value is None:
                if param.get('value', param.get('default', None)):
                    valid_config[param_name] = param.get('value', param.get('default'))
                elif param.get('required'):
                    config_status = 0
            else:
                if same_type:
                    valid_config[param_name] = param_value
                elif param.get('required'):
                    config_status = 0

        config_object['config'] = valid_config
        config_object['status'] = config_status
        config_object['connector'] = conn_id
        if not replace or not config_object.get('config_id', None):
            config_object['config_id'] = str(uuid.uuid4())
    return configuration


def schema_validator(schema, input):
    missing_required_keys = []
    for param in schema:
        if param.get('required', False):
            param_name = param.get('name', '')
            param_value = input.get(param_name, None)
            param_title = param.get('title', param_name)
            if param_value or isinstance(param_value, int):
                if isinstance(param_value, str):
                    if not param_value.strip():
                        missing_required_keys.append(param_title)
            else:
                missing_required_keys.append(param_title)
    return missing_required_keys


def is_operation_param_valid(operation_schema=None, params={}):
    param_schema = operation_schema.get('parameters', None)
    if not params:
        params = {}
    if param_schema:
        if not isinstance(params, dict):
            raise ConnectorError(
                "Invalid Inputs :: Params provided for a action are of invalid format, Expected in dictionary format")

        missing_required_title = schema_validator(param_schema, params)
        if len(missing_required_title):
            raise ConnectorError(
                "Invalid params provided :: Parameters {0} have either blank value or not provided".format(
                    ', '.join(missing_required_title)))
    return params


def validate_config(config, config_schema={}):
    if not config_schema and config:
        error_message = "Invalid configuration :: You are trying to insert configuration for the connector with no connfiguration schema"
        logger.error(error_message)
        raise ConnectorError(error_message)

    missing_required_keys = schema_validator(config_schema.get('fields', []), config)

    if len(missing_required_keys):
        error_message = "Invalid configuration params provided :: Parameters {0} have either blank value or not provided".format(
            ', '.join(missing_required_keys))
        logger.error(error_message)
        raise ConnectorError(error_message)


def is_active_primary():
    get_non_self_primary = make_request('/api/auth/cluster/?self=false&primary=true', 'GET')
    if not get_non_self_primary:
        return True, ''
    nodes = get_non_self_primary['nodes']
    if len(nodes) > 0:
        return False, nodes[0]['nodeId']
    return True, ''


def import_playbook(name, label, version, image, override_playbook_info=True):
    playbooks_path = os.path.join(
        get_connector_path(name, version),
        'playbooks',
        'playbooks.json'
    )
    if not os.path.exists(playbooks_path):
        logger.warn('No playbooks to import.')
        return []
    try:
        with open(playbooks_path) as json_data:
            playbook_data = json.load(json_data)
            playbook_exported_tags = playbook_data.get('exported_tags', [])
            playbook_data = playbook_data.get('data')
    except:
        logger.warn('%s Playbook Import: Error while reading playbook.json.', name)
        return
    # upload image for PB collection
    try:
        if image:
            image_path = os.path.join(get_connector_path(name, version), 'images', image)
            response = make_file_upload_request(image,
                                                open(image_path, 'rb'),
                                                image.replace(' ', '').split('.')[-1],
                                                'images'
                                                )
            for el in playbook_data:
                el['image'] = response.get('@id')
    except Exception as e:
        logger.warn('%s Playbook Import: Error while uploading image for playbook collection %s', name, str(e))

    # Standardise Playbook collection name and description
    if len(playbook_data) == 1 and override_playbook_info:
        playbook_data[0]['name'] = 'Sample - %s - %s' % (label, version)
        playbook_data[0][
            "description"] = 'Sample playbooks for "%s" connector. If you are planning to use any of the sample playbooks in your environment, ensure that you clone those playbooks and move them to a different collection, since the sample playbook collection gets deleted during connector upgrade and delete.' % label

    if len(playbook_exported_tags):
        playbook_tags = []
        for each_tag in playbook_exported_tags:
            playbook_tags.append({'uuid': each_tag})
        try:
            make_request('/api/3/bulkupsert/tags', 'POST', {"__data": playbook_tags, "__unique": ["uuid"]})
        except Exception as e:
            logger.warn('%s Playbook Tags Import: Error while importing playbook tags: %s', name, str(e))

    result = []
    try:
        playbook_collection = make_request('/api/3/', 'POST', playbook_data)
        result = [{'name': item.get('name'),
                   'description': item.get('description'),
                   '@id': item.get('@id'),
                   'image': item.get('image')
                   } for item in playbook_collection.get('persisted', [])
                  ]
    except Exception as e:
        logger.warn('%s Playbook Import: Error while importing playbooks: %s', name, str(e))
    return result


def check_broadcast_to_cluster():
    rpm_installed = False
    propagated = False
    cluster_command = None
    try:
        for parent in psutil.Process().parents():
            try:
                cmdline = parent.cmdline()
                if '/bin/yum' in cmdline:
                    rpm_installed = True
                    if 'cs-propagated' in cmdline:
                        propagated = True
                    else:
                        cluster_command = cmdline[1:]
                        if not '-y' in cluster_command:
                            cluster_command.insert(2, '-y')
                        cluster_command = ' '.join(cluster_command) + ' "cs-propagated"'
                        break
            except Exception as e:
                logger.exception('Error fetching process details')
    except Exception as e:
        logger.exception('Error fetching process info')
    return rpm_installed, propagated, cluster_command


def get_response_primary(primaryNodeId, requestURI, method='GET', data=None):
    curl_command = 'curl -X {0} -k https://localhost:9595{1}'.format(method, requestURI)
    if data:
        curl_command += ' -H "Accept: application/json" -H "Content-Type: application/json" -d \'' + data + '\''
    response = subprocess.check_output(['sudo', '/bin/hamanager', 'add-command', '--nodeId', primaryNodeId,
                                        '--commandId', 'check-health', '--timeout_in_mins', '1', '--command', curl_command])
    response_json = json.loads(response.decode('utf-8').strip('\n'))
    connector_response_json = json.loads(response_json['result'])
    return connector_response_json


def get_truncate_message(value, chr_limit):
    if not type(value) == str:
        value = str(value)
    if len(value) > chr_limit:
        return value[-chr_limit:]
    return value


def connector_actions_serializer(connector, connector_fields):
    connector_object = {}
    for field in connector_fields:
        connector_object[field] = getattr(connector, field, None)
    return connector_object
