import tempfile
import glob
import tarfile
import json
import re
import os
import uuid
import shutil
import copy
import subprocess
from threading import Thread
from django.http import HttpResponse, Http404
from django.http import HttpRequest
from django.conf import settings
from django.db.models import Q
from django_filters.rest_framework import DjangoFilterBackend
from rest_framework import status, filters
from rest_framework.mixins import UpdateModelMixin, DestroyModelMixin, RetrieveModelMixin
from rest_framework.parsers import FileUploadParser
from rest_framework.response import Response
from rest_framework.views import APIView
from rest_framework.viewsets import GenericViewSet, ReadOnlyModelViewSet, ModelViewSet
from connectors.core.connector import logger, ConnectorError
from connectors.decorators import json_response
from connectors.serializers import ConnectorDetailSerializer, ConnectorListSerializer, ConnectorOperationSerializer, \
    ConnectorConfigurationSerializer
from connectors.utils import get_connector_version_or_latest, import_playbook
from integrations.crudhub import make_request
from annotation.views import remove_connector_from_annotation
from connectors.apps import configure_proxy
from connectors.errors.error_constants import *
from connectors.core.constants import STATE_AVAILABLE, NOTIFICATION_BASED_INGESTION
from annotation.views import add_connector_in_annotation
from annotation.models import is_valid_annotation_name
from connectors.models import Connector, Operation, Configuration
from connectors.utils import get_connector_path, get_configuration, on_active_status_update, omit_password, \
    on_connector_delete, get_connector_instance, get_operation, is_operation_param_valid, \
    get_previous_version_config, symlink_system_connector, process_images, invalidate_cache, validate_config, \
    encrypt_password, decrypt_password, update_pwd_field, check_broadcast_to_cluster, get_response_primary, \
    is_active_primary, get_truncate_message, connector_actions_serializer


class ConnectorList(ModelViewSet):
    serializer_class = ConnectorListSerializer
    filter_backends = (filters.OrderingFilter,
                       filters.SearchFilter,
                       DjangoFilterBackend,)
    ordering_fields = ('id', 'name', 'label')
    search_fields = ('$id', '$label')
    filter_fields = ('id', 'name', 'category', 'active', 'ingestion_supported')
    ordering = ('id',)

    def get_queryset(self):
        queryset = Connector.objects.all().defer('icon_large', 'configuration', 'playbook_collections')
        if 'configured' in self.request.GET:
            configured = self.request.GET.get('configured')
            if configured.lower() == 'true':
                queryset = self.filter_queryset(queryset=Connector.objects.filter(~Q(config_count=0)))
            elif configured.lower() == 'false':
                queryset = self.filter_queryset(queryset=Connector.objects.filter(config_count=0))
        if 'tags_contians' in self.request.GET:
            tags = self.request.GET.get('tags_contians')
            queryset = Connector.objects.filter(tags__contains=tags.split(','))
        return queryset

    def connector_actions(self, request, *args, **kwargs):
        connector_name = self.kwargs.get('connector', None)
        connector_version = self.kwargs.get('version', None)
        operation_name = request.GET.get('operation', None)
        connector_fields = ['name', 'version', 'label', 'icon_small']
        operation_fields = ['operation', 'title', 'visible']
        connector_ids_mapping = {}
        connectors_data = []

        if connector_name and connector_version and operation_name:
            connector = Connector.objects.filter(name=connector_name, version=connector_version).first()
            operation = connector.operations.filter(operation=operation_name).first()
            connector_fields.extend(['icon_large', 'publisher', 'cs_approved'])
            connector_fields.remove('icon_small')
            operation_fields.append('parameters')

            connector_object = connector_actions_serializer(connector, connector_fields)
            operation_object = connector_actions_serializer(operation, operation_fields)
            connector_object['operations'] = []
            connector_object['operations'].append(operation_object)
            config_serializier = ConnectorConfigurationSerializer(connector.configuration.all(), many=True)
            connector_object['configurations'] = config_serializier.data
            return Response({"data": connector_object, "status": "Success"}, status=status.HTTP_200_OK)
        else:
            connectors = Connector.objects.filter(~Q(config_count=0) & Q(active=True))
            for connector in connectors:
                connector_object = connector_actions_serializer(connector, connector_fields)
                connector_object['operations'] = []
                connectors_data.append(connector_object)
                connector_ids_mapping[connector.id] = len(connectors_data) - 1

            connector_ids = connector_ids_mapping.keys()
            operations = Operation.objects.filter(connector__in=connector_ids)
            for operation in operations:
                operation_object = connector_actions_serializer(operation, operation_fields)
                connectors_data[connector_ids_mapping[operation.connector.id]]['operations'].append(operation_object)
            return Response({"data": connectors_data, "status": "Success"}, status=status.HTTP_200_OK)


class ConnectorOperationView(ModelViewSet):
    model = Operation
    queryset = Operation.objects.all()
    serializer_class = ConnectorOperationSerializer
    filter_backends = (filters.OrderingFilter,
                       filters.SearchFilter,
                       DjangoFilterBackend,)
    ordering_fields = ('operation',)
    search_fields = ('$operation',)
    filter_fields = ('id', 'operation',)


class ConnectorConfigurationView(ModelViewSet):
    model = Configuration
    queryset = Configuration.objects.all()
    serializer_class = ConnectorConfigurationSerializer
    filter_backends = (filters.OrderingFilter,
                       filters.SearchFilter,
                       DjangoFilterBackend,)
    ordering_fields = ('config_id',)
    search_fields = ('$name', '$config_id',)
    filter_fields = ('id', 'name', 'config_id', 'connector')
    old_data = {}

    def create(self, request, *args, **kwargs):
        try:
            data = request.data
            conn_obj_list = Connector.objects.filter(id=data.get('connector'))

            if not conn_obj_list and not conn_obj_list[0]:
                raise Connector.DoesNotExist
            else:
                conn_obj = conn_obj_list[0]

            conn_instance = get_connector_instance(conn_obj.name, conn_obj.version, conn_obj)
            # check if response should be fetched from primary
            is_primary, primary_node_id = is_active_primary()
            ingestion_modes = conn_obj.metadata.get('ingestion_modes', [])
            if not is_primary and NOTIFICATION_BASED_INGESTION in ingestion_modes:
                connector_response_json = get_response_primary(primary_node_id, request.META['REQUEST_URI'], 'POST',
                                                               json.dumps(data))
                return Response(connector_response_json, status=status.HTTP_200_OK)

            config = data.get('config', {})
            if not data.get('config_id', None):
                data['config_id'] = str(uuid.uuid4())
            validate_config(config, conn_obj.config_schema)
            decrypted_config = copy.deepcopy(config)
            decrypted_config['config_id'] = data.get('config_id')
            decrypted_config['name'] = data.get('name')
            encrypt_password(conn_obj.config_schema, config)

            if data.get('default', False):
                Configuration.objects.filter(connector=data.get('connector'), default=True).update(default=False)

            configuration_serializer = ConnectorConfigurationSerializer(
                data={
                    'config_id': data.get('config_id'),
                    'name': data.get('name', ''),
                    'default': data.get('default', False),
                    'config': config,
                    'connector': data.get('connector')
                }
            )
            configuration_serializer.is_valid(raise_exception=True)
            self.perform_create(configuration_serializer)

            conn_instance.on_add_config(decrypted_config, conn_obj.active)

            result = {
                'configuration': configuration_serializer.data,
                'config_schema': conn_obj.config_schema}
            conn_obj_list.update(config_count=conn_obj.config_count + 1)
            omit_password(result)
            headers = self.get_success_headers(configuration_serializer.data)

            return Response(result['configuration'], status=status.HTTP_201_CREATED, headers=headers)
        except Connector.DoesNotExist:
            logger.error('Cannot find the connector with connector id {0}'.format(data.get('connector')))
            return Response(
                {'message': 'Cannot find the connector with connector id {0}'.format(data.get('connector'))},
                status=status.HTTP_400_BAD_REQUEST)
        except ConnectorError as e:
            return Response({'message': str(e)}, status=status.HTTP_400_BAD_REQUEST)
        except Exception as e:
            return Response(
                {'message': 'Error occurred while configuring connector {0} v{1} ERROR:: {2}'.format(
                    conn_obj.name, conn_obj.version, str(e))},
                status=status.HTTP_400_BAD_REQUEST)

    def update(self, request, *args, **kwargs):
        try:
            partial = kwargs.pop('partial', False)
            instance = self.get_object()
            data = request.data
            conn_obj = Connector.objects.get(id=data.get('connector'))
            conn_instance = get_connector_instance(conn_obj.name, conn_obj.version, conn_obj)

            # check if response should be fetched from primary
            is_primary, primary_node_id = is_active_primary()
            ingestion_modes = conn_obj.metadata.get('ingestion_modes', [])
            if not is_primary and NOTIFICATION_BASED_INGESTION in ingestion_modes:
                connector_response_json = get_response_primary(primary_node_id, request.META['REQUEST_URI'], 'PUT',
                                                               json.dumps(data))
                return Response(connector_response_json, status=status.HTTP_200_OK)

            config = data.get('config', {})
            update_pwd_field(config, instance.config, conn_obj.config_schema)
            validate_config(config, conn_obj.config_schema)

            decrypted_config = copy.deepcopy(config)
            decrypted_config['config_id'] = data.get('config_id')
            decrypted_config['name'] = data.get('name')

            prev_config = instance.config
            decrypt_password(prev_config, conn_obj.config_schema)
            prev_config['config_id'] = instance.config_id
            prev_config['name'] = instance.name

            encrypt_password(conn_obj.config_schema, config)
            data['status'] = 1

            if data.get('default', False):
                Configuration.objects.filter(connector=data.get('connector'), default=True).update(default=False)

            serializer = self.get_serializer(instance, data=data, partial=partial)
            serializer.is_valid(raise_exception=True)
            self.perform_update(serializer)

            if getattr(instance, '_prefetched_objects_cache', None):
                # If 'prefetch_related' has been applied to a queryset, we need to
                # forcibly invalidate the prefetch cache on the instance.
                instance._prefetched_objects_cache = {}

            conn_instance.on_update_config(prev_config, decrypted_config, conn_obj.active)

            result = {
                'configuration': serializer.data,
                'config_schema': conn_obj.config_schema}
            omit_password(result)

            return Response(result['configuration'])

        except Connector.DoesNotExist:
            logger.error('Could not find the connector with connector id {0}'.format(data.get('connector')))
            return Response(
                {'message': 'Could not find the connector with connector id {0}'.format(data.get('connector'))},
                status=status.HTTP_400_BAD_REQUEST)
        except ConnectorError as e:
            return Response({'message': str(e)}, status=status.HTTP_400_BAD_REQUEST)
        except Exception as e:
            return Response({'message': str(e)}, status=status.HTTP_400_BAD_REQUEST)

    def destroy(self, request, *args, **kwargs):
        try:
            instance = self.get_object()
            config = instance.config
            conn_obj_list = Connector.objects.filter(id=instance.connector.id)
            if not conn_obj_list and not conn_obj_list[0]:
                raise Connector.DoesNotExist
            else:
                conn_obj = conn_obj_list[0]

            conn_instance = get_connector_instance(conn_obj.name, conn_obj.version, conn_obj)
            # check if response should be fetched from primary
            is_primary, primary_node_id = is_active_primary()
            ingestion_modes = conn_obj.metadata.get('ingestion_modes', [])
            if not is_primary and NOTIFICATION_BASED_INGESTION in ingestion_modes:
                connector_response_json = get_response_primary(primary_node_id, request.META['REQUEST_URI'], 'DELETE')
                return Response(connector_response_json, status=status.HTTP_200_OK)

            decrypt_password(config, conn_obj.config_schema)
            config['config_id'] = instance.config_id
            config['name'] = instance.name

            self.perform_destroy(instance)

            conn_instance.on_delete_config(config)
            conn_obj_list.update(config_count=conn_obj.config_count - 1)
            return Response(status=status.HTTP_204_NO_CONTENT)

        except Connector.DoesNotExist:
            logger.error('Cannot find the connector with connector id {0}'.format(instance.connector.id))
            return Response(
                {'message': 'Cannot find the connector with connector id {0}'.format(instance.connector.id)},
                status=status.HTTP_400_BAD_REQUEST)
        except ConnectorError as e:
            return Response({'message': str(e)}, status=status.HTTP_400_BAD_REQUEST)
        except Exception as e:
            message = 'Error occurred while deleting the config ERROR :: {0}'.format(str(e))
            logger.error(message)
            return Response({'message': message}, status=status.HTTP_400_BAD_REQUEST)


class ProxySettingView(APIView):
    @json_response
    def post(self, request):
        logger.info('Setting Proxy variables in env')
        configure_proxy(request.data)
        return Response()


class ConnectorInstall(APIView):
    @json_response
    def post(self, request):
        rpm_name = request.data.pop('rpm_name', None)
        connector = request.data.get('name', None)
        version = request.data.get('version', None)
        if rpm_name:
            try:
                if not connector or not version:
                    return Response({'message': 'Invalid Input :: connector name or version was not provided.'},
                                    status=status.HTTP_400_BAD_REQUEST)
                if 'cyops-connector' not in rpm_name:
                    return Response({'message': 'Invalid Input :: rpm name provided was not for connector'},
                                    status=status.HTTP_400_BAD_REQUEST)

                conn_instance = Connector.objects.filter(name=connector, version=version).first()
                if conn_instance and conn_instance.status == 'In-Progress':
                    message = 'Connector you are trying to install is already in progress state Connector :: {0} v{1}'.format(
                        connector, version)
                    return Response({'message': message})
                elif conn_instance and conn_instance.status == 'Completed':
                    message = 'Connector you are trying to install is already installed Connector :: {0} v{1}'.format(
                        connector, version)
                    return Response({'message': message})

                conn_data = request.data
                conn_data['status'] = 'In-Progress'
                serializer = ConnectorDetailSerializer(conn_instance, data=conn_data, partial=True)
                serializer.is_valid(raise_exception=True)
                conn_instance = serializer.save()

                process = Thread(target=install_connector, args=[rpm_name, conn_instance.id])
                process.start()

                return Response({'message': 'Installing connector {0} version {1}'.format(connector, version)})
            except Exception as e:
                logger.error('Error while installing the connector rpm Error :: {0}'.format(str(e)))
        else:
            return Response({'message': 'Invalid Input :: No connector rpm name was provided'},
                            status=status.HTTP_400_BAD_REQUEST)

    def put(self, request):
        return self.post(request)

    def delete(self, request):
        rpm_name = request.data.pop('rpm_name', None)
        connector = request.data.get('name', None)
        version = request.data.get('version', None)
        if rpm_name:
            try:
                if not connector or not version:
                    return Response({'message': 'Invalid Input :: connector name or version was not provided.'},
                                    status=status.HTTP_400_BAD_REQUEST)
                if 'cyops-connector' not in rpm_name:
                    return Response({'message': 'Invalid Input :: rpm name provided was not for connector'},
                                    status=status.HTTP_400_BAD_REQUEST)

                conn_instance = Connector.objects.filter(name=connector, version=version).first()
                if conn_instance and conn_instance.status == 'Uninstallation in progress':
                    message = 'Connector you are trying to uninstall is already in Uninstallation in progress state. Connector :: {0} v{1}'.format(
                        connector, version)
                    return Response({'message': message})

                elif conn_instance and conn_instance.status == 'Uninstalled':
                    message = 'Connector that you are trying to uninstall is already being uninstalled. Connector :: {0} v{1}'.format(
                        connector, version)
                    return Response({'message': message})

                conn_data = request.data
                conn_data['status'] = 'Uninstallation in progress'
                serializer = ConnectorDetailSerializer(conn_instance, data=conn_data, partial=True)
                serializer.is_valid(raise_exception=True)
                conn_instance = serializer.save()

                process = Thread(target=install_connector, args=[rpm_name, conn_instance.id, 'remove'])
                process.start()

                return Response({'message': 'Uninstalling connector {0} version {1}'.format(connector, version)})
            except Exception as e:
                logger.error('Error while uninstalling the connector rpm Error :: {0}'.format(str(e)))
        else:
            return Response({'message': 'Invalid Input :: No connector rpm name was provided'},
                            status=status.HTTP_400_BAD_REQUEST)


def install_connector(rpm_name, conn_id, command='install'):
    try:
        status = 'In-Progress'
        install_result = {}

        def get_truncate(value, chr_limit):
            if not type(value) == str:
                value = str(value)
            if len(value) > chr_limit:
                return value[-chr_limit:]
            return value

        my_env = os.environ.copy()
        my_env['PATH'] = '/usr/sbin:/sbin:' + my_env['PATH']
        install_process = subprocess.Popen(['sudo', '/bin/yum', command, str(rpm_name), '-y'], stdout=subprocess.PIPE,
                                           env=my_env)
        output, err = install_process.communicate(timeout=600)
        if err:
            raise Exception(err)
        elif output and command == 'install':
            try:
                subprocess.check_output('rpm -qia | grep ' + str(rpm_name), shell=True)
            except subprocess.CalledProcessError:
                raise Exception(output)
            install_result['message'] = 'Success'
            install_result['output'] = get_truncate(output, 2000)
            logger.info(
                'Connector {0} {1} command executed successfully. OUTPUT :: {2}'.format(rpm_name, command.title(),
                                                                                        output))
    except subprocess.TimeoutExpired:
        conn_object = Connector.objects.filter(id=conn_id)
        if not conn_object[0].status == 'Completed':
            logger.error('Connector {0} Timed Out {1}'.format(command.title(), rpm_name))
            install_result['message'] = 'Connector {0} Timed Out'.format(command.title())
            install_result['output'] = ''
            status = 'Failed'
        else:
            logger.warn("Connector {0} subprocess got timeout before completing rpm process".format(command))
    except Exception as err:
        logger.error('Connector {0} Failed With ERROR :: {1}'.format(command.title(), str(err)))
        install_result['message'] = 'Connector Installation Failed'
        install_result['output'] = get_truncate(err, 2000)
        status = 'Failed'
    if command == 'install' and status == 'Failed':
        conn_object = Connector.objects.filter(id=conn_id)
        conn_object.update(status=status, install_result=install_result)


class ConnectorImport(APIView):
    parser_classes = (FileUploadParser,)

    @json_response
    def post(self, request, filename):
        logger.info('Importing connector.')
        replace = False
        replace_str = request.GET.get('replace')
        rpm_installed = request.GET.get('rpm_installed')
        if replace_str and (replace_str.lower() == 'true'):
            replace = True
        if rpm_installed:
            if rpm_installed.lower() == 'true':
                rpm_installed = True
            else:
                rpm_installed = False

        try:
            uploaded = request.FILES['file']
            filename = uploaded.name
        except Exception as exp:
            logger.error('{0} ERROR :: {1}'.format(cs_integration_1, str(exp)))
            return Response({'message': cs_integration_1}, status=status.HTTP_400_BAD_REQUEST)

        try:
            # generate a hash and write the file to disk
            connectors_dir = os.path.join(tempfile.gettempdir(), 'connectors')
            if not os.path.exists(connectors_dir):
                os.mkdir(connectors_dir, 0o755)
            connector = os.path.join(connectors_dir, 'conn_%s%s' % (
                filename, settings.CONNECTORS_EXTENSION))
            with open(connector, 'wb') as f:
                for line in uploaded.readlines():
                    f.write(line)
            # Import connector
            # Delete uploaded file from REST call
            delete_uploaded_file = True
            result = import_connector(connector_path=connector,
                                      replace=replace,
                                      delete_uploaded_file=delete_uploaded_file,
                                      rpm_installed=rpm_installed)
            return Response(result)
        except ConnectorError as e:
            logger.exception('from the connector import execption handeling %s' % str(e))
            return Response({'message': str(e)}, status=status.HTTP_400_BAD_REQUEST)
        except Exception as e:
            logger.exception('Error occurred while ')
            return Response({'message': str(e)}, status=status.HTTP_400_BAD_REQUEST)


class ConnectorDetail(RetrieveModelMixin,
                      UpdateModelMixin,
                      DestroyModelMixin,
                      GenericViewSet):
    queryset = Connector.objects.all()
    serializer_class = ConnectorDetailSerializer

    def get_object(self):
        pk = self.kwargs.get('pk', None)
        if pk is not None:
            return super(ConnectorDetail, self).get_object()
        name = self.kwargs.get('name', None)
        version = self.kwargs.get('version', 'latest')
        try:
            version = get_connector_version_or_latest(name, version, None)
            connector = Connector.objects.get(name=name, version=version)
        except Connector.DoesNotExist:
            raise Http404('No matching connector by name %s exists.' % name)
        return connector

    def update(self, request, *args, **kwargs):
        try:
            partial = kwargs.pop('partial', False)
            instance = self.get_object()
            data = request.data
            old_status = instance.active
            new_status = data.get('active', instance.active)
            ingestion_modes = instance.metadata.get('ingestion_modes', [])
            if old_status != new_status:
                # check if response should be fetched from primary
                is_primary, primaryNodeId = is_active_primary()
                if not is_primary and NOTIFICATION_BASED_INGESTION in ingestion_modes:
                    connector_response_json = get_response_primary(primaryNodeId, request.META['REQUEST_URI'], 'PUT',
                                                                   json.dumps(data))
                    return Response(connector_response_json, status=status.HTTP_200_OK)

            serializer = self.get_serializer(instance, data=data, partial=partial)
            serializer.is_valid(raise_exception=True)
            self.perform_update(serializer)

            if old_status != new_status:
                config = get_configuration(
                    **{'conn_id': instance.id, 'decrypt': True, 'config_schema': instance.config_schema,
                       'convert_to_dict': True})
                on_active_status_update(instance.name, instance.version, old_status, new_status, config)
            result = serializer.data
            omit_password(result)
            return Response(result)
        except Exception as e:
            logger.exception('Error occurred while updating the connector ERROR :: {0}'.format(str(e)))
            return Response({'message': 'Error occurred while updating the connector, please check the logs'},
                            status=status.HTTP_400_BAD_REQUEST)

    def help(self, request, *args, **kwargs):
        instance = self.get_object()
        help_path = instance.help_path

        try:
            with open(help_path, 'rb') as pdf:
                response = HttpResponse(pdf.read())
                response['content_type'] = 'application/pdf'
                response['Content-Disposition'] = 'attachment; filename=%s' % os.path.basename(help_path)
                return response
        except Exception as e:
            return Response({'message': str(e)}, status=status.HTTP_400_BAD_REQUEST)

    def retrieve(self, request, *args, **kwargs):
        try:
            instance = self.get_object()
            serializer = self.get_serializer(instance)
            result = serializer.data
            # Omit password field from configuration
            omit_password(result)
        except Http404 as err:
            logger.error(err)
            return Response({'message': str(err)},
                            status=status.HTTP_404_NOT_FOUND
                            )
        except Exception as exp:
            logger.exception('Retrieve:')
            return Response({'message': str(exp)},
                            status=status.HTTP_400_BAD_REQUEST
                            )
        return Response(result)

    @json_response
    def destroy(self, request, *args, **kwargs):
        # ToDo revisit
        try:
            remove_rpm = self.kwargs.pop('remove_rpm', True)
            instance = self.get_object()
            metadata = instance.metadata
            rpm_name = 'cyops-connector-' + instance.name + '-' + instance.version
            remove_rpm = get_remove_rpm(remove_rpm, rpm_name)
            if bool(instance.system) and not self.kwargs.get('system_delete', False):
                return Response({
                    'message': 'Reserved connector \'%s\' cannot be deleted.' % instance.name},
                    status=status.HTTP_400_BAD_REQUEST)

            is_rpm_command, propagated, cluster_command = check_broadcast_to_cluster()
            is_rpm_command = self.kwargs.get('is_rpm_command', is_rpm_command)

            logger.info('is_rpm_command %s, propagated: %s, cluster_command %s', is_rpm_command, propagated,
                        cluster_command)

            if not is_rpm_command and remove_rpm:
                request = HttpRequest()
                request.data = {'name': instance.name, 'version': instance.version, 'rpm_name': rpm_name}
                return ConnectorInstall().delete(request)

            try:
                configuration = get_configuration(
                    **{'conn_id': instance.id, 'decrypt': True, 'config_schema': instance.config_schema,
                       'convert_to_dict': True})
                on_connector_delete(instance.name, instance.version, configuration)
            except Exception as e:
                logger.warn('Error : %s', str(e))

            if propagated:
                path = get_connector_path(instance.name, instance.version)
                if os.path.exists(path):
                    shutil.rmtree(path)
                else:
                    logger.warn('Connector %s folder is not found', instance.name)

                if not is_rpm_command and remove_rpm:
                    delete_thread = Thread(target=remove_connector_rpm, args=[instance.name, instance.version])
                    delete_thread.start()
                return

            try:
                [make_request(collection.get('@id'), 'DELETE', settings.APPLIANCE_PUBLIC_KEY)
                 for collection in instance.playbook_collections]
                [make_request(collection.get('image'), 'DELETE', settings.APPLIANCE_PUBLIC_KEY)
                 for collection in instance.playbook_collections if collection.get('image')]
            except Exception as e:
                logger.warn('Error while deleting associated Playbook Collection : %s', str(e))

            try:
                operation_object = Operation.objects.filter(connector=instance.id).first()
                remove_connector_from_annotation(instance.id, operation_object.operation)
            except Exception as e:
                logger.warn('Error while removing connector %s from annotation(s): %s', instance.name, str(e))

            path = get_connector_path(instance.name, instance.version)
            if os.path.exists(path):
                shutil.rmtree(path)
            else:
                logger.warn('Connector %s folder is not found', instance.name)

            # for system connectors, there could be a dangling symlink
            # but since system connectors can only be replaced, and never deleted
            # the symlink_system_connector will delete the old link
            # remove symlink if system connector delete is supported in future

            self.perform_destroy(instance)

            metadata = instance.metadata
            if metadata.get('rpm_installed', False) and not is_rpm_command:
                delete_thread = Thread(target=remove_connector_rpm, args=[instance.name, instance.version])
                delete_thread.start()

            # publish as a cluster command:
            try:
                if cluster_command:
                    add_command_response = make_request('/api/auth/clustercommand/', 'POST',
                                                        {'commandId': 'connector-remove',
                                                         'command': cluster_command})
                    logger.info('result for publish of connector command: %s' % add_command_response)
            except Exception as e:
                raise ConnectorError('Failed to publish connector delete to cluster')

            return Response(status=status.HTTP_204_NO_CONTENT)

        except Exception as e:
            logger.exception('Destroy:')
            return Response({'message': str(e)}, status=status.HTTP_400_BAD_REQUEST)


def get_remove_rpm(remove_rpm, rpm_name):
    if remove_rpm:
        try:
            output = subprocess.check_output('rpm -qia | grep ' + str(rpm_name), shell=True)
            if rpm_name not in str(output):
                remove_rpm = False
        except subprocess.CalledProcessError:
            remove_rpm = False
    return remove_rpm


def remove_connector_rpm(connector_name, connector_version):
    rpm_name = 'cyops-connector-' + connector_name + '-' + connector_version
    try:
        remove_process = subprocess.Popen(['sudo', '/bin/yum', 'remove', str(rpm_name), '-y'], stdout=subprocess.PIPE)
        output, err = remove_process.communicate()
        if err:
            raise Exception(err)
        elif output:
            logger.info(
                'Connector {0} v{1} rpm removed successfully OUTPUT :: {2}'.format(connector_name, connector_version,
                                                                                   output))
    except Exception as err:
        logger.error('Error while removing the connector rpm Error :: {0}'.format(str(err)))


class ConnectorHealth(APIView):
    @json_response
    def get(self, request, name, version):
        try:
            logger.info('checking connector health')
            config_id = request.GET.get('config', 'get_default_config')
            if not config_id:
                config_id = 'get_default_config'
            conn_obj = Connector.objects.get(name=name, version=version)

            config = get_configuration(**{'conn_id': conn_obj.id, 'config_id': config_id, 'decrypt': True,
                                          'config_schema': conn_obj.config_schema})
            connector = get_connector_instance(name, version, conn_obj)
            # check if response should be fetched from primary
            is_primary, primaryNodeId = is_active_primary()
            ingestion_modes = conn_obj.metadata.get('ingestion_modes', [])
            if not is_primary and NOTIFICATION_BASED_INGESTION in ingestion_modes:
                connector_response_json = get_response_primary(primaryNodeId, request.META['REQUEST_URI'])
                return Response(connector_response_json, status=status.HTTP_200_OK)
            response = connector.verify_health(config,
                                               conn_obj.active)
            response['name'] = name
            response['version'] = version
            return Response(response, status=status.HTTP_200_OK)
        except Connector.DoesNotExist:
            logger.error('Connnector {0} with version {1} not found'.format(name, version))
            return Response({'name': name, 'version': version,
                             'message': 'Could not find connector with the specified id or name',
                             'status': ''}, status=status.HTTP_400_BAD_REQUEST)
        except ConnectorError as e:
            return Response({'name': name, 'version': version,
                             'message': str(e),
                             'status': ''}, status=status.HTTP_400_BAD_REQUEST)
        except Exception as e:
            message = 'Error on health check of the connector {0} version {1}'.format(name, version)
            logger.exception('{0} ERROR {1}'.format(message, str(e)))
            return Response({'name': name, 'version': version,
                             'message': str(e),
                             'status': ''}, status=status.HTTP_400_BAD_REQUEST)


class ConnectorExecute(APIView):

    @staticmethod
    def execute_connector_operation(input_data, name=None, version=None):
        if not (name or version):
            name = input_data.get('connector')
            version = get_connector_version_or_latest(name, input_data.get('version'))
        input_data['version'] = version
        operation = input_data.get('operation', '')
        params = input_data.get('params', {})
        env = input_data.get('env', {})
        input_data.pop('env', None)
        config_id = input_data.get('config', 'get_default_config')

        if not config_id:
            config_id = 'get_default_config'
        conn_obj = Connector.objects.get(name=name, version=version)

        if conn_obj.active:
            connector = get_connector_instance(name, version, conn_obj)

            config = get_configuration(**{'conn_id': conn_obj.id,
                                          'config_id': config_id,
                                          'decrypt': True, 'config_schema': conn_obj.config_schema})

            if config and not config.get('status', 0):
                message = 'The config {0} is partially configured and cannot be used for execution. ' \
                          'Complete and validate the connector configuration ' \
                          'by providing required inputs for the missing fields in the configuration.'

                logger.error(message.format(config.get('name')))
                raise ConnectorError(message.format(config.get('name')))

            operation = get_operation(operation, conn_obj.id)
            params = is_operation_param_valid(operation, params)
            input_data['config'] = config
            input_data['params'] = params
            # change the handle operation params as there will no input validate check

            if input_data.get('audit', False):
                audit_data = {
                    'connector': conn_obj.label,
                    'action': operation.get('title'),
                    'config': config.get('name'),
                    'version': version,
                    'params': params
                }
                input_data['audit_info']['data'] = audit_data

            response, is_binary = connector._handle_operation(input_data,
                                                              env=env
                                                              )
            return response, is_binary
        else:
            raise ConnectorError(cs_integration_4)

    @json_response
    def post(self, request):
        input_data = request.data
        name = input_data.get('connector')
        version = get_connector_version_or_latest(name, input_data.get('version'))
        
        try:
            response, is_binary = ConnectorExecute.execute_connector_operation(input_data, name, version)
            if is_binary:
                return HttpResponse(content=response, content_type='application/octet-stream',
                                    status=status.HTTP_200_OK)
            else:
                return Response(response, status=status.HTTP_200_OK)

        except Connector.DoesNotExist:
            logger.error('Connnector {0} with version {1} not found'.format(name, version))
            return Response({'name': name, 'version': version,
                             'message': 'Could not find connector with the specified id or name',
                             'status': ''}, status=status.HTTP_400_BAD_REQUEST)

        except ConnectorError as e:
            message = '{0}  Connector :: {1}V{2}'.format(str(e), name, version)
            logger.error(message)
            return Response({'message': message}, status=status.HTTP_400_BAD_REQUEST)

        except Exception as exp:
            message = '{0} ERROR :: {1}'.format(cs_integration_5, str(exp))
            logger.exception(message)
            return Response(
                {'message': message},
                status=status.HTTP_400_BAD_REQUEST
            )


def import_connector(connector_path, replace, isReserved=False, delete_uploaded_file=False, *args, **kwargs):
    # if its a cluster propagation command, db entry is not to be updated
    # check if there is a cluster command to install this connector
    rpm_installed, propagated, cluster_command = check_broadcast_to_cluster()
    rpm_installed = kwargs.get('rpm_installed', rpm_installed)
    if propagated:
        info = get_file_info(connector_path, replace, isReserved)
        install_connector_files(connector_path, info.get('name'), info.get('version'), replace)
        return ''

    # import previous version config if config schema is same
    info = validate_info(connector_path, replace, isReserved)
    try:
        config = get_previous_version_config(info, replace)
    except Exception as e:
        config = dict()
        logger.warn('{0} ERROR :: {1}'.format(cs_integration_2, str(e)))

    # If Replace, remove all existing connectors
    if replace:
        connectors = Connector.objects.filter(name=info.get('name'), status='Completed')
        for connector in connectors:
            remove_rpm = True
            if info.get('version') == connector.version:
                remove_rpm = False
            conn_detail_view = ConnectorDetail()
            conn_detail_view.kwargs = {'name': connector.name, 'version': connector.version, 'system_delete': True,
                                       'remove_rpm': remove_rpm, 'is_rpm_command': rpm_installed}
            conn_detail_view.destroy({})

    install_connector_files(connector_path, info.get('name'), info.get('version'), replace)

    playbook_collections = import_playbook(
        info.get('name'),
        info.get('label'),
        info.get('version'),
        info.get('icon_large_name'),
        info.get('override_playbook_info', True)
    )
    try:
        result = insert_connector(info=info, playbook_collections=playbook_collections, config=config,
                                  isReserved=isReserved, rpm_installed=rpm_installed)
    except ConnectorError as e:
        if delete_uploaded_file: os.unlink(connector_path)
        raise ConnectorError(e)

    if delete_uploaded_file: os.unlink(connector_path)

    # publish as a cluster command:
    try:
        if cluster_command:
            add_command_response = make_request('/api/auth/clustercommand/', 'POST',
                                                {'commandId': 'connector-install', 'command': cluster_command})
            logger.info('result for publish of connector command: %s' % add_command_response)
    except Exception as e:
        raise ConnectorError('Failed to publish connector install to cluster')
    return result


def install_connector_files(connector_path, name, version, replace=True):
    # remove previous folder if replace is true
    if replace:
        try:
            installed_connector_path = get_connector_path(name, '*')
            for path in glob.glob(installed_connector_path):
                shutil.rmtree(path)
        except Exception as e:
            logger.warn('Error while deleting connector: %s folder.', name)

    # extract it to connectors.
    tar = tarfile.open('%s' % (connector_path), 'r:gz')
    tar_extract_dir = os.path.join(settings.CONNECTORS_DIR, 'temp')
    os.makedirs(tar_extract_dir, exist_ok=True)
    tar.extractall(path=tar_extract_dir)
    tar.close()
    # rename version to folder.
    os.rename(os.path.join(tar_extract_dir, name),
              get_connector_path(name, version))

    # create symlink for system connector to latest version
    if name in settings.CONNECTORS_RESERVED:
        symlink_system_connector(name, version)


def update_connector(connector, configurations={}):
    name = connector.get('name')
    version = connector.get('version')
    system = connector.get('system')
    playbook_collections = connector.get('playbook_collections')
    new_configurations = []
    connector_path = get_connector_path(name, version)
    info_path = os.path.join(connector_path, 'info.json')
    try:
        with open(info_path) as json_file:
            info = json.load(json_file)
            if not info.get('name') == name or not info.get('version') == version:
                logger.error('Name or version mismatched in info file and database')
                raise ConnectorError('Name or version mismatched in info file and database')
    except Exception as e:
        logger.error('Error occurred while retrieving the info json ERROR :: {0}'.format(str(e)))
        raise ConnectorError('Error occurred while retrieving the info json ERROR :: {0}'.format(str(e)))

    if configurations:
        for config_id, config in configurations.items():
            config['config_id'] = config_id
            config = {
                'config_id': config_id,
                'name': config.pop('name'),
                'default': config.pop('default', False),
                'config': config
            }
            new_configurations.append(config)

    insert_connector(info, playbook_collections, new_configurations, isReserved=system)


def insert_connector(info, playbook_collections=[], config={}, isReserved=False, rpm_installed=False):
    # validate file and delete if invalid.
    connector_update = False
    try:
        conn_instance = Connector.objects.filter(name=info.get('name', None), version=info.get('version', None)).first()
        if conn_instance:
            connector_update = True
        else:
            conn_instance = None
        icon_small, icon_large = process_images(info)
        connector_fields = [f.name for f in Connector._meta.get_fields()]
        connector_data = {}
        info_copy = copy.deepcopy(info)
        for field in connector_fields:
            value = info.get(field, None)
            if value or isinstance(value, bool):
                if field == 'configuration':
                    connector_data['config_schema'] = value
                elif field == 'operations':
                    continue
                else:
                    connector_data[field] = value
                info_copy.pop(field, None)
            elif field == 'configuration':
                connector_data['config_schema'] = {}
                connector_data['config_count'] = -1

        connector_data['system'] = isReserved
        connector_data['icon_small'] = icon_small
        connector_data['icon_large'] = icon_large
        connector_data['playbook_collections'] = playbook_collections
        connector_data['metadata'] = info_copy
        connector_data['metadata']['rpm_installed'] = rpm_installed
        connector_data['migrate'] = True
        connector_data['active'] = True
        connector_data['status'] = 'Completed'
        connector_data['install_result'] = {}
        connector_data['install_result']['message'] = 'Success'
        connector_data['install_result']['output'] = ''

        if connector_update and conn_instance:
            serializer = ConnectorDetailSerializer(conn_instance, data=connector_data, partial=True)
        else:
            serializer = ConnectorDetailSerializer(data=connector_data)
        serializer.is_valid(raise_exception=True)
        conn_instance = serializer.save()

        if bool(config):
            if isinstance(config, list):
                config_count = len(config)
                for each_config in config:
                    each_config['connector'] = conn_instance.id
                configuration_serializer = ConnectorConfigurationSerializer(data=config, many=True)
            else:
                config_count = 1
                configuration_serializer = ConnectorConfigurationSerializer(
                    data={
                        'config_id': config.get(''),
                        'name': config.get('name', ''),
                        'default': config.get('default', False),
                        'config': config,
                        'connector': conn_instance.id
                    }
                )
            configuration_serializer.is_valid(raise_exception=True)
            config_instance = configuration_serializer.save()
            serializer.update(conn_instance, {'config_count': conn_instance.config_count + config_count})

        operations = []
        operation_fields = [f.name for f in Operation._meta.get_fields()]
        for operation in info_copy.get('operations', []):
            operation_object = {'connector': conn_instance.id}
            for field in operation_fields:
                value = operation.get(field, None)
                if value or isinstance(value, bool):
                    operation_object[field] = value
                    operation.pop(field, None)
            operation_object['metadata'] = operation
            operations.append(operation_object)

        if operations:
            operation_serializer = ConnectorOperationSerializer(data=operations, many=True)
            operation_serializer.is_valid(raise_exception=True)
            operation_serializer.save()

            add_connector_in_annotation((serializer.data).get('id'), info)

        result = serializer.data
        result['data'] = info
        omit_password(result)
        invalidate_cache()

        return result

    except ConnectorError as e:
        raise ConnectorError(str(e))

    except Exception as exp:
        # clean up when went wrong.
        if conn_instance and connector_update:
            install_result = {}
            install_result['message'] = 'Connector Insertion Failed'
            install_result['output'] = get_truncate_message(exp, 2000)
            serializer.update(conn_instance, {'status': 'Failed', 'install_result': install_result})
        elif conn_instance:
            conn_instance.delete()
        # clean up when went wrong.
        try:
            shutil.rmtree(
                get_connector_path(info.get('name'), info.get('version')))
        except:
            pass
        logger.exception('{0} ERROR :: {1}'.format(cs_integration_3, str(exp)).format(info.get('name')))
        raise ConnectorError('{0} ERROR :: {1}'.format(cs_integration_3, str(exp)).format(info.get('name')))


log_dir = os.path.dirname(settings.APP_LOG_FILE_PATH)


def get_file_info(filename, replace, isReserved=False):
    try:
        with tarfile.open(filename) as td:
            contents = td.getnames()
    except:
        raise ConnectorError(
            'Error reading Adapter package. File must be in .tgz format.')

    infonames = [fn for fn in contents if fn.endswith('/info.json')]
    if not len(infonames) == 1:
        raise ConnectorError(
            'Invalid connector structure :: info file of connector are either multiple or not provided.')

    conn_files = [fn for fn in contents if fn.endswith('/connector.py')]
    if not len(conn_files) == 1:
        raise ConnectorError(
            'Invalid connector structure :: connector.py file of connector either multiple or not provided.')

    try:
        with tarfile.open(filename, encoding='utf-8') as td:
            f = td.extractfile(infonames[0])
            info = json.loads(f.read().decode('utf-8'))
    except Exception as exp:
        raise ConnectorError('Error occurred while reading info file, inappropriate format for info block. ERROR :: {0}'
                             .format(str(exp)))

    name = info.get('name', '')
    version = info.get('version', '')

    if not re.compile(r'[0-9]+.[0-9]+.[0-9]+$').match(version):
        raise ConnectorError('Invalid Inputs :: version field either is of invalid format or not provided.')

    if not re.compile(r'[a-zA-Z_][a-zA-Z0-9_]*').match(name):
        raise ConnectorError('Invalid Inputs :: name field either is of invalid format or not provided.')

    if name + '/info.json' != infonames[0]:
        raise ConnectorError(
            'Name mismatch :: connector folder name and name in info.json should be same.')

    if name in ('core',) or (not isReserved and name in settings.CONNECTORS_RESERVED):
        raise ConnectorError('Invalid inputs :: Connector name provided is a reserved name')

    if not replace and os.path.exists(
            get_connector_path(name, version)):
        raise ConnectorError('Connector conflict :: Connector with same name is already active')
    return info


def validate_info(filename, replace, isReserved=False):
    info = get_file_info(filename, replace, isReserved)
    # validate annotation
    annotation_category_map = {}
    for op in info.get('operations'):
        if op.get('annotation'):
            if op.get('annotation') in annotation_category_map and \
                            op.get('category', 'miscellaneous').lower() != annotation_category_map.get(
                        op.get('annotation')):
                raise ConnectorError(
                    'Same annotation %s is defined in categories %s and %s. An annotation can be defined for only one \
                    category considering category as miscellaneous if not defined.' % (
                        op.get('annotation'),
                        op.get('category', 'miscellaneous'),
                        annotation_category_map.get(op.get('annotation'))
                    )
                )
            annotation_category_map.update({op.get('annotation'): op.get('category', 'miscellaneous').lower()})
            is_valid_annotation_name(op.get('annotation').lower(), op.get('category'))

    config_names = [each_field.get('name') for each_field in info.get('configuration', {}).get('fields', [])]
    if 'default' in config_names:
        logger.error('default is a reserved key in the connector config to specify whether a config is default or not. \
        It cannot be used in connector configuration schema.')
        raise ConnectorError('default is a reserved key in the connector config to specify whether a config is default or not. \
        It cannot be used in connector configuration schema.')

    if 'name' in config_names:
        logger.error('name is a reserved key in the connector config to specify whether a config is default or not. \
        It cannot be used in connector configuration schema.')
        raise ConnectorError('name is a reserved key in the connector config to specify whether a config is default or not. \
        It cannot be used in connector configuration schema.')

    return info


def install_connector_packages(connector_name, connector_version):
    file = get_connector_path(connector_name, connector_version)
    file = os.path.join(file, 'requirements.txt')
    command = ['sudo', '-u', 'nginx', '/opt/cyops-integrations/.env/bin/pip', 'install', '-r', file]
    my_env = os.environ.copy()
    my_env['PATH'] = '/usr/sbin:/sbin:' + my_env['PATH']
    try:
        install_process = subprocess.Popen(command, stdout=subprocess.PIPE, stderr=subprocess.PIPE, env=my_env)
        output, err = install_process.communicate()
        if err:
            raise Exception(err)
        elif output:
            logger.info('All the packages for connector installed successfully. OUTPUT :: {0}'.format(output))
    except Exception as e:
        logger.exception('Error in installing the packages of connector {0} v{1} ERROR:: {2}'.format(connector_name,
                                                                                                     connector_version,
                                                                                                     str(e)))
