import ast
import json
import requests
import logging
from django.conf import settings
from cshmac.requests import HmacAuth
from .requests_auth import get_requests_auth

logger = logging.getLogger('connectors')


def make_request(url, method, body=None, *args, **kwargs):
    """
    This function facilitates using the crud hub api.

    It is for general purpose requests, but takes care of authentication
    automatically.

   :param str collection: An IRI that points to the location of the \
       crud hub collection (E.g. /api/3/events)
   :param str method: HTTP method
   :param dict body: An object to json encode and send to crud hub

   :return: the API response either as a json-like dict if possible or as bytes
   :rtype: dict/bytes
    """
    # ctrl c/v
    # get rid of the body on GET/HEAD requests
    bodyless_methods = ['head', 'get']
    if method.lower() in bodyless_methods:
        body = None

    if type(body) == str:
        try:
            body = ast.literal_eval(body)
        except Exception:
            pass

    url = settings.CRUD_HUB_URL + url
    logger.info('Starting request: %s , %s', method, url)
    env = kwargs.get('env', {})
    # for default if no env HMAC method
    if not env or not env.get('auth_info', False):
        env['auth_info'] = {"auth_method": "CS HMAC"}

    # for public and private key in env
    if env.get('public_key', False) and env.get('private_key', False):
        env['auth_info'] = {"auth_method": "CS HMAC"}
        public_key = env.get('public_key')
        private_key = env.get('private_key')
    else:
        public_key = settings.APPLIANCE_PUBLIC_KEY
        private_key = settings.APPLIANCE_PRIVATE_KEY
    auth_info = env.get('auth_info')
    auth = get_requests_auth(auth_info, url, method,
                             public_key,
                             private_key,
                             json.dumps(body), *args, **kwargs)

    response = requests.request(method, url, auth=auth, json=body, verify=False)
    return maybe_json_or_raise(response)


def make_file_upload_request(file_name, file_content, file_type, collection='files', *args, **kwargs):
    url = settings.CRUD_HUB_URL + '/api/3/{0}'.format(collection)
    auth = HmacAuth(url, 'POST', settings.APPLIANCE_PUBLIC_KEY,
                    settings.APPLIANCE_PRIVATE_KEY,
                    settings.APPLIANCE_PUBLIC_KEY.encode('utf-8'))
    files = {'file': (file_name, file_content, file_type, {'Expire': 0})}
    response = requests.post(url, auth=auth, files=files, verify=False)
    return maybe_json_or_raise(response)


def maybe_json_or_raise(response):
    """
    Helper function for processing request responses

    Returns any json found in the response. Otherwise, it will extract the
    response as text, or, failing that, as bytes.

    :return: the response from the request
    :rtype: dict or str or bytes
    :raises: :class:`requests.HTTPError` if status code was 4xx or 5xx
    """
    if response.ok:
        try:
            logger.info('Processing request responses.')
            return response.json(strict=False)
        except Exception:
            return response.text or response.content
    else:
        msg = ''
        try:
            msg = response.json()
            logger.warn(msg)
        except Exception:
            pass
        if not msg:
            msg = response.text
            logger.warn(msg)

        try:
            response.raise_for_status()
        except requests.exceptions.HTTPError as e:
            # add any response content to the error message
            msg = '{} :: {} :: Url: {}'.format(str(e), msg, response.url)
            logger.error(msg)
            raise requests.exceptions.HTTPError(msg, response=response)
